#ifndef BEEP_POLYPHONY_H
#define BEEP_POLYPHONY_H

#include <cstdlib>
#include <algorithm>
#include <sys/types.h>
#include <sys/time.h>

#include "common.h"

template< typename Beep, typename Envelope1, typename Envelope2, typename Envelope3, unsigned int slot_count >
class VirtualPolyphony {
  public:
    typedef unsigned int SlotID;
    VirtualPolyphony( const char *_device_name ) : beep( _device_name ), current_time( 0.0 ) {
      SlotID slot;
      for( slot = 0; slot != slot_count; slot++ ) {
        if( slot_map[ slot ].state != NOTE_IDLE ) {
          slot_map[ slot ].state = NOTE_IDLE;
        }
      }

    }
    SlotID noteOn( int _new_tone, unsigned int _which_envelope ) {
      SlotID new_slot = getSlot();
      slot_map[ new_slot ].state = NOTE_ON;
      slot_map[ new_slot ].note = _new_tone;
      if( _which_envelope == 0 )
        slot_map[ new_slot ].envelope = envelope1 + new_slot;
      else if( _which_envelope == 1 )
        slot_map[ new_slot ].envelope = envelope2 + new_slot;
      else if( _which_envelope == 2 )
        slot_map[ new_slot ].envelope = envelope3 + new_slot;
      slot_map[ new_slot ].envelope->noteOn();
      slot_map[ new_slot ].note_on_time = current_time;
      setLevel( new_slot, slot_map[ new_slot ].envelope->get( 0.0f ) );
      return new_slot;
    }
    void noteOff( SlotID _id ) {
      slot_map[ _id ].state = NOTE_OFF;
      slot_map[ _id ].envelope->noteOff( current_time - slot_map[ _id ].note_on_time );
    }
    inline double getTime() const {
      struct timeval tv;
      gettimeofday( &tv, NULL );
      return static_cast< double >( tv.tv_sec ) + static_cast< double >( tv.tv_usec ) * 0.000001;
    }
    inline void sleep( double _until ) const {
      long int sleep_time = static_cast< long int >( ( _until - getTime() ) * 1000000 );
      usleep( sleep_time );
    }
    void operator()( float _length ) {
      play( _length );
    }
  private:
    void calcEnvelope() {
      SlotID slot;
      for( slot = 0; slot != slot_count; slot++ ) {
        if( slot_map[ slot ].state != NOTE_IDLE ) {
          setLevel( slot, slot_map[ slot ].envelope->get( current_time - slot_map[ slot ].note_on_time ) );
        }
      }
    }
    void play( float _length ) {
      float total_level = getTotalLevel();
      float total_time = std::max( 1.0f, total_level );
      SlotID slot;
      for( slot = 0; slot != slot_count; slot++ ) {
        if( slot_map[ slot ].state != NOTE_IDLE ) {
          double play_time = _length * ( slot_map[ slot ].level / total_time );
          double time = beep.getTime();
          if( slot_map[ slot ].note ) {
            beep.setTone( slot_map[ slot ].note );
            setLevel( slot, slot_map[ slot ].envelope->get( current_time - slot_map[ slot ].note_on_time ) );
            if( slot_map[ slot ].level == 0.0f ) {
              slot_map[ slot ].state = NOTE_IDLE;
            }
            beep.sleep( time + play_time );
          }
          else {
            setLevel( slot, slot_map[ slot ].envelope->get( current_time - slot_map[ slot ].note_on_time ) );
            if( slot_map[ slot ].level == 0.0f ) {
              slot_map[ slot ].state = NOTE_IDLE;
            }
            beep.noize( time + play_time );
          }
        }
      }
      if( total_time > total_level ) {
        beep.setTone( 0 );
        double time = beep.getTime();
        beep.sleep( time + ( total_time - total_level ) * _length );
      }
      current_time += _length;
    }
    float getTotalLevel() {
      SlotID slot;
      float level = 0.0f;
      for( slot = 0; slot != slot_count; slot++ ) {
        if( slot_map[ slot ].state != NOTE_IDLE ) {
          level += slot_map[ slot ].level;
        }
      }
      return level;
    }
    void setLevel( SlotID _id, float _level ) {
      if( _level != 0.0f ) {
        slot_map[ _id ].level = _level;
      }
      else {
        slot_map[ _id ].state = NOTE_IDLE;
      }
    }
    SlotID getSlot() {
      SlotID mapped_slot;
      SlotID oldest_off_slot = (unsigned int)-1;
      unsigned int oldest_off_note_on_time = (unsigned int)-1;
      SlotID oldest_on_slot = (unsigned int)-1;
      unsigned int oldest_on_note_on_time = (unsigned int)-1;
      for( mapped_slot = 0; mapped_slot != slot_count; mapped_slot++ ) {
        if( slot_map[ mapped_slot ].state == NOTE_IDLE ) {
          return mapped_slot;
        }
        else if( slot_map[ mapped_slot ].state == NOTE_OFF ) {
          if( oldest_off_note_on_time > slot_map[ mapped_slot ].note_on_time ) {
            oldest_off_note_on_time = slot_map[ mapped_slot ].note_on_time;
            oldest_off_slot = mapped_slot;
          }
        }
        else {
          if( oldest_on_note_on_time > slot_map[ mapped_slot ].note_on_time ) {
            oldest_on_note_on_time = slot_map[ mapped_slot ].note_on_time;
            oldest_on_slot = mapped_slot;
          }
        }
      }
      if( oldest_off_slot != (unsigned int)-1 )
        return oldest_off_slot;
      if( oldest_on_slot != (unsigned int)-1 )
        return oldest_on_slot;
      return 0;
    }
    Beep beep;
    Envelope1 envelope1[ slot_count ];
    Envelope2 envelope2[ slot_count ];
    Envelope3 envelope3[ slot_count ];
    SlotState slot_map[ slot_count ];
    int current_tone;
    int console;
    double current_time;
};

#endif
